/* Multiple-file auto-sequence finder.
   Copyright (C) 1997 Free Software Foundation, Inc.
   Franois Pinard <pinard@iro.umontreal.ca>, 1997.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "system.h"
#include "getopt.h"

/* Variables.  */

/* The name this program was run with. */
const char *program_name;

/* If nonzero, display usage information and exit.  */
static int show_help = 0;

/* If nonzero, print the version on standard output and exit.  */
static int show_version = 0;

/* Minimum size of a cluster, not counting ignored lines.  */
int threshold_size = 4;

/* Give statistics operation statistics.  */
int statistics = 0;

/* Show progress of operations.  */
int verbose = 0;

/* Line descriptions.  */

/* Comparison work is done on checksums of lines, rather than input lines
   themselves.  The following array maintains one such checksum for each
   line, whether real or sentinel.  Checksums are computed so they ignore
   parts of the line which are not really meaningful for the purpose of
   comparisons.  If a line has all its parts ignored, the IGNORABLE flag is
   set, and for some algorithms, this is a little like if the line has never
   been part of its input file.  If the SENTINEL flag is set, the line may
   not compare equally to any other line.  If either flag is set, the
   checksum is not meaningful, and is kept zero for helping later sorts.  */

struct line_object
{
  int sentinel : 1;
  int ignorable : 1;
  int checksum : 30;
};

static struct line_object *line_object = NULL;
static int line_objects = 0;

/*-----------------------------------------------------------------------.
| Move line pointer forward or backward, skipping over ignorable lines.	 |
`-----------------------------------------------------------------------*/

#define FORWARD_LINE(Pointer) \
  do { Pointer++; } while (Pointer->ignorable)

#define BACKWARD_LINE(Pointer) \
  do { Pointer--; } while (Pointer->ignorable)

/*---------------------------------.
| Incrementally compute checksum.  |
`---------------------------------*/

#define ADJUST_CHECKSUM(Checksum, Character) \
  (Checksum = (Checksum << 5) + (Checksum >> 25) + (Character & 0xFF))

/*-----------------.
| Add a new line.  |
`-----------------*/

static inline void
add_line (int sentinel, int ignorable, int checksum)
{
  struct line_object *line;

  if (line_objects % 1024 == 0)
    line_object = (struct line_object *)
      xrealloc (line_object,
		(line_objects + 1024) * sizeof (struct line_object));

  line = line_object + line_objects++;
  line->sentinel = sentinel;
  line->ignorable = ignorable;
  line->checksum = checksum;
}

/*---------------------.
| Add a new sentinel.  |
`---------------------*/

static void
add_sentinel (void)
{
  add_line (1, 0, 0);
}

/*-------------------------------------------------------------------------.
| Find how many input lines, including embedded ignored lines, are part of |
| a run of lines known to have a number of non-ignored lines.		   |
`-------------------------------------------------------------------------*/

static int
real_size (int first, int number)
{
  struct line_object *line = line_object + first;

  while (number-- > 0)
    {
      line++;

      if (number > 0)
	while (line->ignorable)
	  line++;
    }
  return line - line_object - first;
}

/*------------------------------------------------------------------.
| Find how many identical lines follow, given two start lines.  The |
| starting lines may not be sentinels or ignored lines.		    |
`------------------------------------------------------------------*/

static int
identical_size (int index1, int index2)
{
  struct line_object *line1 = line_object + index1;
  struct line_object *line2 = line_object + index2;
  int result = 0;

  while (line1->checksum == line2->checksum)
    {
      result++;

      FORWARD_LINE (line1);
      FORWARD_LINE (line2);

      if (line1->sentinel || line2->sentinel)
	break;
    }
  return result;
}

/* File descriptions.  */

/* All input files use a single continuous line counting scheme.  To revert
   line numbers back to actual file and line references, the following array
   maintains the cumulative number of lines seen at end of each file, for
   each file given as argument to the program.  To ease some algorithms, a
   few sentinel lines exist: one before the first file, one between each
   file, and one after the last file.  Each file also counts all sentinels
   preceding it.  */

struct input_object
{
  const char *file_name;
  int cumulative_lines;
};

static struct input_object *input_object = NULL;
static int input_objects = 0;

/*---------------------------------------------------------.
| Read a file, given its NAME, and construct descriptors.  |
`---------------------------------------------------------*/

static void
add_file (const char *name)
{
  struct input_object *input;
  FILE *file;			/* file being read */
  int character;		/* character being seen */
  unsigned int checksum;	/* line checksum */
  int empty;			/* if line is empty */
  int line_count;		/* number of lines read */

  /* Read the file and checksum all lines.  For now, ignore all prefix
     whitespace in lines.  */

  if (file = fopen (name, "r"), !file)
    error (EXIT_FAILURE, errno, name);
  if (verbose)
    fprintf (stderr, _("Reading %s ..."), name);

  empty = 1;
  checksum = 0;
  line_count = 0;
  while (character = getc (file), character != EOF)
    switch (character)
      {
      default:
	ADJUST_CHECKSUM (checksum, character);
	empty = 0;
	break;

      case ' ':
      case '\t':
	if (!empty)
	  ADJUST_CHECKSUM (checksum, character);
	break;

      case '\n':
	add_line (0, empty, checksum);
	line_count++;
	empty = 1;
	checksum = 0;
	break;
      }
  if (!empty)
    {
      add_line (0, 0, checksum);
      line_count++;
    }

  if (verbose)
    fprintf (stderr, _(" %d lines\n"), line_count);

  fclose (file);

  /* Add a new input file descriptor.  */

  if (input_objects % 16 == 0)
    input_object = (struct input_object *)
      xrealloc (input_object,
		(input_objects + 16) * sizeof (struct input_object));

  input = input_object + input_objects++;
  input->file_name = name;
  input->cumulative_lines = line_objects;
}

/*-----------------------------------------------------------------.
| Output the file and line reference for an internal line number.  |
`-----------------------------------------------------------------*/

static void
output_reference (int number)
{
  struct input_object *input;
  int preceding_count = 0;

  for (input = input_object;
       number > input->cumulative_lines;
       input++)
    preceding_count = input->cumulative_lines;

#if 0
  int lowest;			/* current lowest possible index */
  int highest;			/* current highest possible index */
  int middle;			/* current middle index */

  lowest = 0;
  highest = input_objects - 1;
  while (lowest <= highest)
    {
      middle = (lowest + highest) / 2;
      input = input_object + middle;
      if (number < *input)
	highest = middle;
      else if (number > *input)
	lowest = middle + 1;
    }
#endif

  fputs (input->file_name, stdout);
  printf (":%d: ", number - preceding_count);
}

/* Groups of identical lines.  */

/* For finding identical clusters, the program uses the following indirect
   array, which is merely a list of indices into the line_object array,
   sorted in such a way to bring together lines having identical checksums.
   Whenever two checksums are equal, lines are then ordered by the checksum
   of the following lines, and this way forward until a difference is found.
   Sentinels go last, and compare in a stable way.  Ignorable lines are
   never used in comparisons.  */

static int *indirect_line;
static int indirect_lines;

/*----------------------------.
| Compare two lines indices.  |
`----------------------------*/

static int
compare_indices (const void *void_first, const void *void_second)
{
#define value1 *((int *) void_first)
#define value2 *((int *) void_second)
  struct line_object *line1 = line_object + value1;
  struct line_object *line2 = line_object + value2;
  int result;

  while (1)
    {
      /* Order checksums.  */

      result = line1->checksum - line2->checksum;
      if (result)
	return result;

      /* Seek backward for a difference.  */

      FORWARD_LINE (line1);
      FORWARD_LINE (line2);

      /* Sentinels are never equal (unless really the same).  They go after
	 all checksums, and compare so to stabilize the sort.  */

      if (line1->sentinel)
	return line2->sentinel ? value1 - value2 : 1;

      if (line2->sentinel)
	return -1;
    }
}

/*---------------------------------.
| Create the indirect line array.  |
`---------------------------------*/

static void
create_indirect_array (void)
{
  int counter;

  indirect_line = (int *) xmalloc (line_objects * sizeof (int));
  indirect_lines = 0;
  for (counter = 0; counter < line_objects; counter++)
    if (!line_object[counter].sentinel
	&& !line_object[counter].ignorable)
      indirect_line[indirect_lines++] = counter;
  if (indirect_lines < line_objects)
    indirect_line = (int *)
      xrealloc (indirect_line, indirect_lines * sizeof (int));

  if (indirect_lines > 1)
    qsort (indirect_line, indirect_lines, sizeof (int), compare_indices);
}

/* Clusters and members.  */

/* A cluster represents one set of lines which is repeated at two or more
   places in the input files, while each repetition is called a cluster
   member.  The number of members is the number of repetitions, and the
   number of lines does not take ignorable lines into account.  */

struct cluster_object
{
  short member_count;
  short line_count;
};

static struct cluster_object *cluster_object = NULL;
int cluster_objects = 0;

/* A cluster member starts at some line number and runs for a given number
   of lines.  Ignorable lines may be embedded in a cluster member, so
   cluster members of a single cluster may have different lengths, but a
   cluster member never starts nor end with an ignorable line.  Also, of
   course, sentinels are never part of any cluster.  By construction,
   all cluster members of a same cluster are next to each other.  */

struct member_object
{
  short cluster_number;
  int first_line;
};

static struct member_object *member_object = NULL;
static int member_objects = 0;

/*--------------------.
| Add a new cluster.  |
`--------------------*/

static inline void
add_cluster (int line_count)
{
  struct cluster_object *cluster;

  if (cluster_objects % 8 == 0)
    cluster_object = (struct cluster_object *)
      xrealloc (cluster_object,
		(cluster_objects + 8) * sizeof (struct cluster_object));

  cluster = cluster_object + cluster_objects++;
  cluster->member_count = 0;
  cluster->line_count = line_count;
}

/*---------------------------.
| Add a new cluster member.  |
`---------------------------*/

static inline void
add_member (int line)
{
  struct member_object *member;

  if (member_objects % 8 == 0)
    member_object = (struct member_object *)
      xrealloc (member_object,
		(member_objects + 8) * sizeof (struct member_object));

  member = member_object + member_objects++;
  member->cluster_number = cluster_objects - 1;
  member->first_line = line;
  cluster_object[cluster_objects - 1].member_count++;
}

/*----------------------.
| Search for clusters.  |
`----------------------*/

static void
find_clusters (void)
{
  int *start;			/* start of current cluster */
  int *limit;			/* one past limit for current cluster */
  int *cursor;			/* cursor from start to limit */
  struct line_object *line;	/* line_object cursor */
  int checksum;			/* cheksum under consideration */
  int cluster_size;		/* size of current cluster */

  for (start = indirect_line;
       start + 1 < indirect_line + indirect_lines;
       start = limit)
    {
      limit = start + 1;
      cluster_size = identical_size (*start, *limit);

      /* A cluster with a single member is no cluster at all.  */
      if (cluster_size < threshold_size)
	continue;

      /* Find all members of the current cluster.  */
      limit++;
      while (limit < indirect_line + indirect_lines)
	{
	  int size = identical_size (*start, *limit);

	  if (size < threshold_size)
	    break;

	  if (size < cluster_size)
	    cluster_size = size;
	  limit++;
	}

      /* If all members are preceded by identical lines, then they are
	 indeed part of a bigger cluster, which has been or will be caught
	 elsewhere.  Then, it's no use retaining them as separate, here.  */
      line = line_object + *start;
      BACKWARD_LINE (line);
      if (!line->sentinel)
	{
	  checksum = line->checksum;
	  for (cursor = start + 1; cursor < limit; cursor++)
	    {
	      line = line_object + *cursor;
	      BACKWARD_LINE (line);
	      if (line->sentinel || line->checksum != checksum)
		break;
	    }
	  if (cursor == limit)
	    continue;
	}

      /* Register all found clusters.  */
      add_cluster (cluster_size);
      for (cursor = start; cursor < limit; cursor++)
	add_member (*cursor);
    }
}

/*------------------------------------------------------------------------.
| Sort helper for clusters.  Arrange so bigger clusters appear first, and |
| use original input order afterwards.					  |
`------------------------------------------------------------------------*/

static int
compare_members (const void *void_first, const void *void_second)
{
#define member1 ((struct member_object *) void_first)
#define member2 ((struct member_object *) void_second)
  int result;

  result = (cluster_object[member2->cluster_number].line_count
	    - cluster_object[member1->cluster_number].line_count);
  if (result)
    return result;

  result = member1->cluster_number - member2->cluster_number;
  if (result)
    return result;

  return member1->first_line - member2->first_line;
}

/*------------------------.
| Output found clusters.  |
`------------------------*/

static void
output_clusters (void)
{
  struct member_object *member;
  struct cluster_object *cluster;
  int position;

  qsort (member_object, member_objects, sizeof (struct member_object),
	 compare_members);

  for (member = member_object;
       member < member_object + member_objects;
       member++)
    {
      if (member == member_object ||
	  member->cluster_number != (member - 1)->cluster_number)
	position = 0;

      output_reference (member->first_line);
      cluster = cluster_object + member->cluster_number;
      printf ("Member %d/%d of cluster %d -- %d lines\n",
	      ++position,
	      cluster->member_count,
	      member->cluster_number + 1,
	      real_size (member->first_line, cluster->line_count));
    }
}

/* Main control.  */

/*-----------------------------------------------.
| Explain how to use the program, then get out.	 |
`-----------------------------------------------*/

static void
usage (int status)
{
  if (status != EXIT_SUCCESS)
    fprintf (stderr, _("Try `%s --help' for more information.\n"),
	     program_name);
  else
    {
      printf ("GNU %s %s\n", PACKAGE, VERSION);
      printf (_("\
\n\
Usage: %s [OPTION]... FILE...\n"), program_name);
      fputs (_("\
\n\
  -t, --threshold=LINES   do not consider clusters of less than LINES lines\n\
  -s, --statistics        produce statistics on stdout\n\
  -v, --verbose           report progress of operations on stderr\n\
      --help              display this help and exit\n\
      --version           output version information and exit\n"),
	     stdout);
      fputs (_("\
\n\
Report bugs to <wdiff-bugs@gnu.ai.mit.edu>.\n"),
	       stdout);
    }
  exit (status);
}

/*----------------------------------------------------------------------.
| Main program.  Decode ARGC arguments passed through the ARGV array of |
| strings, then launch execution.				        |
`----------------------------------------------------------------------*/

/* Long options equivalences.  */
static const struct option long_options[] =
{
  {"help", no_argument, &show_help, 1},
  {"statistics", no_argument, NULL, 's'},
  {"threshold", required_argument, NULL, 't'},
  {"verbose", no_argument, NULL, 'v'},
  {"version", no_argument, &show_version, 1},
  {0, 0, 0, 0},
};

int
main (int argc, char *const *argv)
{
  int option_char;		/* option character */

  program_name = argv[0];
  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);

  /* Decode command options.  */

  if (argc == 1)
    usage (EXIT_SUCCESS);

  while (option_char = getopt_long (argc, argv, "st:v", long_options, NULL),
	 option_char != EOF)
    switch (option_char)
      {
      default:
	usage (EXIT_FAILURE);

      case '\0':
	break;

      case 's':
	statistics = 1;
	break;

      case 't':
	threshold_size = atoi (optarg);
	break;

      case 'v':
	verbose = 1;
	break;
      }

  /* Process trivial options.  */

  if (show_version)
    {
      printf ("GNU %s %s\n", PACKAGE, VERSION);
      fputs (_("\
\n\
Copyright (C) 1997 Free Software Foundation, Inc.\n"),
	     stdout);
      fputs (_("\
This is free software; see the source for copying conditions.  There is NO\n\
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
	     stdout);
      fputs (_("\
\n\
Written by Franc,ois Pinard <pinard@iro.umontreal.ca>.\n"),
	     stdout);
      exit (EXIT_SUCCESS);
    }

  if (show_help)
    usage (EXIT_SUCCESS);

  add_sentinel ();
  while (optind < argc)
    {
      add_file (argv[optind++]);
      add_sentinel ();
    }
  if (statistics)
    printf (_("Read summary: %d files, %d lines\n"),
	    input_objects, line_objects - input_objects - 1);

  if (verbose)
    fprintf (stderr, _("Sorting ..."));
  create_indirect_array ();
  if (verbose)
    fprintf (stderr, _(" done\n"));

  if (verbose)
    fprintf (stderr, _("Clustering ..."));
  find_clusters ();
  if (verbose)
    fprintf (stderr, _(" done\n"));

  if (statistics)
    {
      int count = 0;
      struct cluster_object *cluster;

      for (cluster = cluster_object;
	   cluster < cluster_object + cluster_objects;
	   cluster++)
	count += cluster->line_count * (cluster->member_count - 1);

      printf (_("Work summary: %d clusters, %d members, %d duplicate lines\n"),
	      cluster_objects, member_objects, count);
    }

  output_clusters ();

  exit (0);
}
